import asyncio
import logging
import time
from functools import lru_cache
from typing import Dict, Iterable, Iterator, List, Set, Tuple, TypeVar, cast

from fastapi import Depends, HTTPException, status
from ldap3.utils.conv import escape_filter_chars

from id_broker_common.kelvin import get_kelvin_obj, kelvin_session
from id_broker_common.pseudonyms_udm_rest import (
    IdBrokerPseudonymisationError,
    get_service_provider_attribute,
    ldap_credentials,
    ldap_settings,
)
from id_broker_common.utils import remove_prefix
from ucsschool.apis.models import OAuth2Token
from ucsschool.apis.utils import LDAPAccess, auth_manager, get_logger
from ucsschool.kelvin.client import SchoolClass, SchoolClassResource, Session, User, UserResource

from .models import ClassMembers, NonEmptyStr, NoStarStr, Student, Teacher
from .settings import PLUGIN_SETTINGS_FILE, kelvin_access_parallelism_setting

T = TypeVar("T")
logger: logging.Logger = get_logger()


# Extracted to variable to allow dep overrides in FastAPI, which cannot resolve anonymous dependencies
# TODO: import PLUGIN_NAME without circular import error:
token_dep = auth_manager("self_disclosure", OAuth2Token)


# {user_name: {ou: set(role, ...)}}
user_name_to_role_per_school_cache: Dict[str, Dict[str, Iterable[str]]] = {}


def service_provider_from_token(token: OAuth2Token = Depends(token_dep)) -> str:
    return token.client_id


def connecting_user_pseudonym_from_token(token: OAuth2Token = Depends(token_dep)) -> str:
    return token.sub


@lru_cache(maxsize=1024)
def role_split(ucsschool_role: str) -> Tuple[str, str, str]:
    """Cached: same ucsschool_role string results in same tuple."""
    role, context_type, context_name = ucsschool_role.split(":")
    return role, context_type, context_name


@lru_cache(maxsize=1024)
def roles_in_ucsschool_roles(ucsschool_roles: Tuple[str, ...], school_name: str) -> Set[str]:
    """Cached: same list of ucsschool_roles strings (hashable->sorted tuple) returns same set of roles."""
    return {
        role
        for role, context_type, context_name in [
            role_split(ucsschool_role) for ucsschool_role in ucsschool_roles
        ]
        if context_type == "school" and school_name == context_name
    }


def roles_in_school(user: User, school_name: str) -> Set[str]:
    """Helper to get only the roles that are defined for a user within a specific school"""
    try:
        return roles_in_ucsschool_roles(tuple(sorted(user.ucsschool_roles)), school_name)
    except ValueError:
        logger.error("Bad value in 'ucsschool_roles' attribute of %r: %r", user, user.ucsschool_roles)
        return set()


async def get_users_roles_in_school(user_name: str, school: str, *, session: Session) -> Iterable[str]:
    """Get the roles of user `user_name` in ou `school`. If not in cache, retrieve it through Kelvin."""
    if school not in user_name_to_role_per_school_cache.get(user_name, {}):
        user = cast(User, await get_kelvin_obj(UserResource(session=session), name=user_name))
        update_user_name_to_role_per_school_cache(user)
    return user_name_to_role_per_school_cache[user_name][school]


def update_user_name_to_role_per_school_cache(user: User) -> Dict[str, Iterable[str]]:
    """Update cache from user data."""
    user_name_to_role_per_school_cache[user.name] = {
        a_school: roles_in_school(user, a_school) for a_school in user.schools
    }
    return user_name_to_role_per_school_cache[user.name]


async def update_missing_users_in_user_name_to_role_per_school_cache_parallel(
    user_names: List[str], school: str, *, session: Session
) -> None:
    """
    Retrieve data through Kelvin for all users in `user_names`, that are missing entries for ou `school`
    in `user_name_to_role_per_school_cache`. Kelvin users are retrieved in parallel.
    """
    if user_names_to_update := [
        user_name
        for user_name in user_names
        if school not in user_name_to_role_per_school_cache.get(user_name, {})
    ]:
        users: List[User] = await retrieve_users_in_parallel(user_names_to_update, session=session)
        for user in users:
            update_user_name_to_role_per_school_cache(user)


def split_list(li: List[T], chunk_size: int) -> Iterator[List[T]]:
    """Split a list into chunks of equal size (except last chunk, which is <= `chunk_size`)."""
    len_li = len(li)
    for i in range(0, len_li, chunk_size):
        yield li[i : i + chunk_size]


async def retrieve_users_in_parallel(user_names: List[str], *, session: Session) -> List[User]:
    """Retrieve multiple Kelvin user objects in parallel."""
    users: List[User] = []
    t0 = time.time()
    parallelism = kelvin_access_parallelism_setting(PLUGIN_SETTINGS_FILE)
    for user_name_list in split_list(user_names, parallelism):
        users.extend(
            await asyncio.gather(
                *[
                    get_kelvin_obj(UserResource(session=session), name=user_name)
                    for user_name in user_name_list
                ]
            )
        )
    t1 = time.time()
    logger.debug(
        "Retrieved %d users (parallelism: %d) from Kelvin in %.3f seconds.",
        len(user_names),
        parallelism,
        t1 - t0,
    )
    return users


async def extract_members(
    schoolclass: SchoolClass,
    service_provider: str,
    session: Session,
    list_students=True,
    list_teachers=True,
) -> ClassMembers:
    """
    Helper to get users of a school_class split into teachers and students

    :raises id_broker_common.udm_rest.IdBrokerPseudonymisationError: if `service_provider` could not be
        loaded from / found in `settings/data` UDM object
    """
    teachers = []
    students = []
    user_name: str
    sp_attribute = await get_service_provider_attribute(service_provider)
    users: List[User] = await retrieve_users_in_parallel(schoolclass.users, session=session)
    for user in users:
        school_authority_prefix = f"{user.source_uid}-"
        if sp_attribute not in user.udm_properties:
            logger.warning(
                f"Ignoring user {user.name!r} without pseudonym in {sp_attribute!r} (for service "
                f"provider {service_provider!r})."
            )
            continue
        roles_in_schools = update_user_name_to_role_per_school_cache(user)
        roles = roles_in_schools[schoolclass.school]
        if list_students and "student" in roles:
            student = Student(
                user_id=user.udm_properties[sp_attribute],
                username=remove_prefix(user.name, school_authority_prefix),
                firstname=user.firstname,
                lastname=user.lastname,
            )
            students.append(student)
        if list_teachers and "teacher" in roles:
            teacher = Teacher(
                user_id=user.udm_properties[sp_attribute],
                username=remove_prefix(user.name, school_authority_prefix),
            )
            teachers.append(teacher)
    return ClassMembers(students=students, teachers=teachers)


async def get_user(pseudonym: str, service_provider: str, session: Session) -> User:
    """
    Retrieve user.

    :param str pseudonym: a users pseudonym
    :param str service_provider: the service provider for which to search user with the given pseudonym
    :param Session session: Kelvin client session instance
    :return: Kelvin User object
    :raises HTTPException: 404 if not found
    :raises id_broker_common.udm_rest.IdBrokerPseudonymisationError: if `service_provider` could not be
        loaded from / found in `settings/data` UDM object
    """
    sp_attribute = await get_service_provider_attribute(service_provider)
    search_params = {sp_attribute: pseudonym}
    async for user in UserResource(session=session).search(**search_params):
        # We expect entryUUID values (the only individual element in a pseudonym) to be unique not only
        # for a domain, but globally. If that'd prove to be wrong, the search parameters should either
        # include the school authority (stored in 'ucsschoolSourceUID') or collect all results of the
        # search and raise an error if more than 1 is found.
        return user
    else:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"No user was found with pseudonym {pseudonym!r} for service provider "
            f"{service_provider!r}.",
        )


async def get_connected_user(
    service_provider: str = Depends(service_provider_from_token),
    pseudonym_of_connecting_user: str = Depends(connecting_user_pseudonym_from_token),
    session: Session = Depends(kelvin_session),
) -> User:
    logger.debug(
        "service_provider=%r pseudonym_of_connecting_user=%r",
        service_provider,
        pseudonym_of_connecting_user,
    )
    try:
        return await get_user(pseudonym_of_connecting_user, service_provider, session)
    except (HTTPException, IdBrokerPseudonymisationError) as exc:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED) from exc


async def get_school_class(pseudonym: str, service_provider: str, session: Session) -> SchoolClass:
    """
    Retrieve school class.

    :param str pseudonym: a school class' pseudonym
    :param str service_provider: the service provider for which to search group with the given pseudonym
    :param Session session: Kelvin client session instance
    :return: Kelvin SchoolClass object
    :raises HTTPException: 404 if not found
    :raises id_broker_common.udm_rest.IdBrokerPseudonymisationError: if `service_provider` could not be
        loaded from / found in `settings/data` UDM object
    """
    sp_attribute = await get_service_provider_attribute(service_provider)
    attrs = await _get_obj_attr_by_pseudonym(pseudonym, sp_attribute, "ucsschoolGroup", "cn")
    school = attrs["school"]
    name = remove_prefix(attrs["name"], f"{school}-")
    return cast(
        SchoolClass,
        await get_kelvin_obj(
            SchoolClassResource(session=session),
            name=name,
            school=school,
        ),
    )


async def _get_obj_attr_by_pseudonym(
    pseudonym: NoStarStr,
    sp_attribute: str,
    object_class: NonEmptyStr,
    name_attr: NonEmptyStr,
) -> Dict[str, str]:
    """Get 'name' and 'school' from `name_attr` and `ucsschoolRole` of LDAP object."""
    ldap = LDAPAccess(ldap_settings(), ldap_credentials())
    filter_s = f"(&(objectClass={object_class})" f"({sp_attribute}={escape_filter_chars(pseudonym)}))"
    results = await ldap.search(filter_s, attributes=[name_attr, "ucsschoolRole"])
    if len(results) == 1:
        result = results[0]
        name = result[name_attr].value
        if isinstance(result.ucsschoolRole.value, str):
            school_value = result.ucsschoolRole.value
        else:
            school_value = [s for s in result.ucsschoolRole.value if ":school:" in s][0]
        school = school_value.split(":")[-1]
        return {"name": name, "school": school}
    elif len(results) > 1:
        logger.error("More than 1 result when searching LDAP with filter %r: %r.", filter_s, results)
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f"Error retrieving object with pseudonym {pseudonym!r}.",
        )
    else:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"Object with pseudonym {pseudonym!r} not found.",
        )
